# Copyright (C) 2010 Anthony D. Castiglia and John T. Whelan
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""The detgeom module contains routines which define and manipulate
the geometry of Earth-based gravitational wave (GW) detectors. They
take as arguments human-readable quantities such as locations
specified by latitude and longitude in decimal degrees and height in
meters above a reference ellipsoid (the WGS-84 Earth model) or
orientations specified by azimuth and altitude angles in decimal
degrees. Major existing GW detectors are pre-defined.  Internally, the
geometry is manipulated via geographical structures which contain
angles (latitude, longitude, azimuth, etc) in radians, as well as
geometrical structures which describe the detector via a position
vector and response tensor in Cartesian coordinates.

detgeom provides the following objects:
    Location : geodetic location 
    Orientation : Earth-based orientation 
    Detector : general GW detector 
    IfoDetector : laser interferometric detector 
    BarDetector : resonant bar detector
    Source : GW source object

and methods: 
    LLO, LHO, etc. : Create IfoDetector or BarDetector
    object for a number of known Earth-based detectors.

    overlapreductionfunction : Return the overlap reduction function
    between two GW detectors for a range of frequencies.

    orfintegrand : Return the integrand of the overlap reduction
    function between two detectors for a range of frequencies.

Also:
    site : Dictionary of known detectors. These can be accessed by
    using detgeom.site['{one-letter site key}']() for IFO detectors,
    and by detgeom.site['{one-letter site key}'](azDeg) for bar
    detectors, where azDeg is the azimuth angle (measured in degrees
    east from North) of the detector.

One-letter site keys correspond to the first letter in the site's
channel prefix (e.g. 'H' in 'H1:LSC-AS_Q' corresponds to 'LHO'). The
correspondence is as specified in the frame format specification,
http://www.ligo.caltech.edu/docs/T/T970130-F.pdf

Dependencies: scipy, numpy

"""
import numpy as np
import scipy.special as sci
import pdb

WGS84_A_SI = 6.378137e+6    # Semimajor axis of WGS-84 Ref Ellipsoid in meters
WGS84_B_SI = 6.356752314e+6 # Semiminor axis of WGS-84 Ref Ellipsoid in meters

# Julian days of inserted leap seconds
leapSecs = np.array([2441317.5, 2441499.5, 2441683.5, 2442048.5, 2442413.5,
                     2442778.5, 2443144.5, 2443509.5, 2443874.5, 2444239.5,
                     2444786.5, 2445151.5, 2445516.5, 2446247.5, 2447161.5,
                     2447892.5, 2448257.5, 2448804.5, 2449169.5, 2449534.5,
                     2450083.5, 2450630.5, 2451179.5, 2453736.5, 2454832.5])

# First possible future leap second
leapSecsEnd = 2455197.5

class Location(object):
    """Geodetic location object
   
    Parameters:
        latDeg : geodetic latitude in degrees
        lonDeg : geodetic longitude in degrees
        h : Elevation above WGS-84 Reference Ellipsoid in meters

    Attributes: 
        lat : geodetic latitude in radians 
        lon : geodetic longitude in radians 
        height : Elevation above WGS-84 Reference Ellipsoid in meters

    Methods: 
        getcartsesianposition : Calculate the cartesian position
        vector from geodetic location.
   
    """
    def __init__(self, latDeg, lonDeg, h = 0.):
        self.lat = latDeg * np.pi / 180.
        self.lon = lonDeg * np.pi / 180.
        self.height = h

    def getcartesianposition(self):
        """Construct Cartesian position vector for Location object.

        getcartesianposition calculates the Cartesian position vector
        r (in meters) corresponding to a point defined by geodetic
        latitude, longitude, and height in a Cartesian coordinate
        system defined by

             i) The intersection of the Equator and Prime Meridian,
            ii) the intersection of the Equator and the meridian at 90
                degrees East longitude, and
           iii) the North Pole.
        
        The position is calculated assuming the latitude, longitude
        and height are ellipsoidal coordinates on an ellipsoid of
        semimajor axis 6.378137e+6 meters and semiminor axis
        6.356752314e+6. These coordinates are described in much
        greater detail in
        http://www.ligo.caltech.edu/docs/T/T980044-10.pdf
        
        See also: Location

        """
        a2 = WGS84_A_SI**2 # square of semimajor axis (equatorial radius)
        b2 = WGS84_B_SI**2 # square of semiminor axis (polar radius)
        coslat = np.cos(self.lat)
        sinlat = np.sin(self.lat)
        R = np.sqrt(a2*coslat**2 + b2*sinlat**2)
        R1 = (a2/R + self.height)*coslat
        R2 = (b2/R + self.height)*sinlat
        return np.array([R1*np.cos(self.lon), R1*np.sin(self.lon), R2])

class Orientation(object):
    """Earth-based orientation object.
    
    Parameters:
        azDeg : azimuth in degrees East (clockwise) of North
        altDeg : altitude (tilt) in degrees above the local tangent
        plane

    Attributes: 
        az : azimuth in radians East (clockwise) of North,
        and alt : altitude (tilt) in radians above the local tangent
        plane,

    Methods: 
        getcartesiandirection : Calculate Cartesian unit vector
        from Earth- based orientation.

    See also: Location, getcartesianposition

    """
    def __init__(self, loc, azDeg, altDeg = 0):
        self.az = azDeg * np.pi / 180.
        self.alt = altDeg * np.pi / 180.
        self.loc = loc # a Location object

    def getcartesiandirection(self):
       """Construct Cartesian unit vector for Orientation object.
       
       getcartesiandirection calculates the unit vector u in
       Earth-based Cartesian coordinates corresponding to an
       orientation described by the azimuth and altitude angles, at a
       location on the Earth described by the latitude, longitude, and
       elevation, in a Cartesian coordinate system defined by
             i) The intersection of the Equator and Prime Meridian,
            ii) the intersection of the Equator and the meridian at 90
                degrees East longitude, and
           iii) the North Pole.
       
       The inputs are the azimuth angle in radians East (clockwise) of
       North and the altitude (tilt) angle in radians above the local
       tangent plane.

       The function first calculates the components of the unit vector
       in a local Cartesian basis (with axes aligned in the local
       East, North, and Up directions), then transforms the vector
       first into an orthonormal basis based on Earth-fixed
       cylindrical coordinates and finally into Earth -fixed Cartesian
       coordinates.

       See also: Orientation

       """
       uloc = np.array([np.cos(self.alt)*np.sin(self.az),
                         np.cos(self.alt)*np.cos(self.az),
                         np.sin(self.alt)])
       coslat = np.cos(self.loc.lat)
       sinlat = np.sin(self.loc.lat)
       coslon = np.cos(self.loc.lon)
       sinlon = np.sin(self.loc.lon)

       # Transformation matrix from E, N, U to cylindrical basis:
       Rcylloc = np.array([[0.,     -sinlat, coslat] , 
                           [1.,      0.    , 0.    ] ,
                           [0.,      coslat, sinlat]])
        
       # Transformation matrix from cylindrical to Cartesian basis:
       Rcarcyl = np.array([[coslon,-sinlon, 0     ] ,
                           [sinlon, coslon, 0     ] ,
                           [0     , 0     , 1     ]])

       Rcarloc = np.dot(Rcarcyl, Rcylloc)
       return np.dot(Rcarloc, uloc)

class Detector(object):
    """Detector object with position r and detector response tensor d.

    Attributes: 
        r : Cartesian position vector in Earth-fixed coordinates
        d : Cartesian detector response tensor
    
    Metods:
        getantennapatternfactors(Source) : Return F-plus and F-cross
        for Detector and GW Source objects.
    
    See also: getantennapatternfactors, IFODetector, BarDetector,
    Source

    """
    def __init__(self, r, d):
        self.r = r
        self.d = d
    
    def amCoefficients(self, skyPos, gstHr):
        """Calculate amplitude moduation coefficients a,b for a given
        sky position.

        amCoefficients(skyPos, gstHr) calculates amplitude modulation
        coefficients a, b for the detector, user-defined sky-position,
        and given GST.

        Parameters:
        skyPos -- detgeom.SkyPosition object
        gstHr -- Greenwich sidereal time in hours
        """
        gamma = gstHr * np.pi / 12. # sideral time in radians
        # Rotation matrix for rotations about e3 axis (passive rotation):

        # Check: if a detector arm has components [1,0,0] in
        # co-rotating coordinates, it should have components
        # [cos(gamma),sin(gamma),0] in intertial coordinates

        R_gamma = np.array([[ np.cos(gamma),-np.sin(gamma), 0.],
                            [ np.sin(gamma), np.cos(gamma), 0.],
                            [ 0.           , 0.           , 1.]])
        # Transform detector tensor from co-rotating to inertial coordinates
        d = np.dot(np.dot(R_gamma, self.d), R_gamma.transpose())

        a = np.dot(skyPos.epsilon_plus,d).trace(0,-2,-1)
        b = np.dot(skyPos.epsilon_cross,d).trace(0,-2,-1)
        
        return (a, b)
    
    def antennaPatterns(self, source, gstHr):
        """Calculate antenna patterns F_+ and F_x for a given source.

        antennaPatterns(source, gstHr) calculates antenna pattern
        factors F_+ and F_x for the detector, user-defined source, and
        given GST.

        Parameters:
        source -- detgeom.Source object
        gstHr -- Greenwich sidereal time in hours
        """

        (a,b) = self.amCoefficients(source.skyPosition,gstHr)

        F_plus = a * np.cos(2 * source.psi) + b * np.sin(2 * source.psi)
        F_cross = -a * np.sin(2 * source.psi) + b * np.cos(2 * source.psi)
        
        return (F_plus, F_cross)

class IFODetector(Detector):
    """Cartesian object describing IFO geometry.
    
    Parameters:
        loc : Location object
        xarm : Orientation object for IFO's x-arm 
        yarm : Orientation object for IFO's y-arm
    
        :or:

        Keyword Arguments:

            latdeg : geodetic latitude in degrees
            lonDeg : geodetic longitude in degrees
            h : elevation above WGS-84 Reference Ellipsoid in meters
            xarmAzDeg : angle of x-arm East of North in degrees
            xarmAltDeg : angle (tilt) between x-arm and local tangent
            plane in degrees
            yarmAzDeg : angle of y-arm East of North in degrees
            yarmAltDeg : angle (tilt) between y-arm and local tangent
            plane in degrees

    IFODetector objects can be created either by specifying Location
    and Orientaiton objects, or by specifying all of the keyword
    arguments listed above.
    
    See also: Detector, Location, Orientation, getcartesianposition,
    getcartesiandirection

    """
    def __init__(self, loc = None, xarm = None, yarm = None, **kwargs):
        if (loc == None):
            self.loc = Location(kwargs.get('latDeg'), kwargs.get('lonDeg'), 
                                kwargs.get('h'))
            self.xarm = Orientation(self.loc, kwargs.get('xarmazDeg'),
                                    kwargs.get('xarmaltDeg'))
            self.yarm = Orientation(self.loc, kwargs.get('yarmazDeg'),
                                    kwargs.get('yarmaltDeg'))
        
        else:
            self.loc = loc
            self.xarm = xarm
            self.yarm = yarm

        r = self.loc.getcartesianposition()
        self.xhat = self.xarm.getcartesiandirection()
        self.yhat = self.yarm.getcartesiandirection()
        d = 0.5 * (np.outer(self.xhat, self.xhat) - np.outer(self.yhat, 
                                                              self.yhat))
        Detector.__init__(self, r, d)
        

class BarDetector(Detector):
    """Cartesian object describing bar geometry.
    
    Parameters:
        loc : Location object
        axis : Orientation object for BarDetector's axis
    
        :or:

        Keyword Arguments:
            latdeg : geodetic latitude in degrees
            lonDeg : geodetic longitude in degrees
            h : elevation above WGS-84 Reference Ellipsoid in meters
            azDeg : angle of detector axis East of North in degrees
            altDeg : angle (tilt) between detector axis and local
            tangent plane in degrees

    BarDetector objects can be created either by specifying Location
    and Orientaiton objects for loc and axis, or by specifying all of
    the keyword arguments listed above.

    See also: Location, Orientation, getcartesianposition,
    getcartesiandirection
    
    """
    def __init__(self, loc=None, axis=None, **kwargs):
        if (loc == None):
            self.loc = Location(kwargs.get('latDeg'), kwargs.get('lonDeg'), 
                                kwargs.get('h'))
            self.axis = Orientation(self.loc, kwargs.get('azDeg'),
                               kwargs.get('altDeg'))
        else:
            self.loc = loc
            self.axis = axis

        r = self.loc.getcartesianposition()
        self.uhat = self.axis.getcartesiandirection()
        d = np.outer(self.uhat, self.uhat)
        Detector.__init__(self, r, d)

class SkyPosition(object):
    """Geometrical sky position object.
    
    Parameters: 
    
        decDeg: Declination in decimal degrees
        raHr: Right ascension in hours sidereal time

    Attributes: 

        dec: Declination in radians
        ra: Right ascension in radians
        i, j, k : orthonormal polarization basis; k in direction of
        source, i parallel to celestial equator, in direction of
        decreasing right ascension, j = k x i (in intertial basis 
        {e1,e2,e3} )
    """
    def __init__(self, decDeg, raHr):
        # Determine shape of input arrays
        try:
            decShape = decDeg.shape
        except AttributeError:
            # If decDeg is a scalar, use a zero-length tuple for its shape
            decShape = ()

        try:
            raShape = raHr.shape
        except AttributeError:
            # If raHr is a scalar, use a zero-length tuple for its shape
            raShape = ()

        if raShape != decShape:
            raise RuntimeError, 'Shape of RA and declination arrays much match'

        # Convert declination from degrees to radians:
        self.dec = decDeg * np.pi / 180.
        # Convert RA from hours to radians:
        self.ra = raHr * np.pi / 12.

        # We want to construct basis vectors with one more index than
        # the input arrays.  We do this by first reshaping unit
        # vectors and trig functions into arrays with the desired
        # numbers of indices:
        uShape = tuple(np.ones(len(decShape),dtype=int)) + (3,)
        aShape = decShape + (1,) # concatenation of tuples

        e1 = np.array([1,0,0]).reshape(uShape)
        e2 = np.array([0,1,0]).reshape(uShape)
        e3 = np.array([0,0,1]).reshape(uShape)

        cosDec = np.cos(self.dec).reshape(aShape)
        sinDec = np.sin(self.dec).reshape(aShape)
        cosRA = np.cos(self.ra).reshape(aShape)
        sinRA = np.sin(self.ra).reshape(aShape)

        # Then we combine them with arithmetic resulting in vectors
        # with the desired shape of decShape + (3,)
        self.k = - cosDec * cosRA * e1 - cosDec * sinRA * e2 - sinDec * e3
        self.i =            sinRA * e1 -          cosRA * e2
        self.j = - sinDec * cosRA * e1 - sinDec * sinRA * e2 + cosDec * e3

        vShape = self.i.shape

        lShape = decShape + (3,1)
        rShape = decShape + (1,3)

        il = self.i.reshape(lShape)
        jl = self.j.reshape(lShape)
        ir = self.i.reshape(rShape)
        jr = self.j.reshape(rShape)

        self.epsilon_plus = il * ir - jl * jr
        self.epsilon_cross = il * jr + jl * ir

    def getAMCoefficients(self, detector, gstHr, *args, **kwargs):
        """Return amplitude modulation coefficients (a, b).

        getAMCoefficients(Det) returns the values for a and b or the
        Detector object specified. The input Det is a Detector object.
        
        Alternatively, a and b may be computed by calling
        Detector.getantennapatternfactors(skyPos), and specifying a
        SkyPosition object.

        Example:

        See also: Detector, IFODetector, BarDetector

        """
        return detector.amCoefficients(self,gstHr)

class Source(object):
    """GW source geometrical object.
    
    Parameters:
       
        skyPosition: skyPosition object with declination and right
        ascension attributes 
        polarizationDeg : Polarization angle in degrees

    Attributes:

        i, j, k : orthonormal polarization basis; k in direction of
        source, i parallel to celestial equator, in direction of
        decreasing right ascension, j = k x i (in intertial basis 
        {e1,e2,e3} )
        epsilon_plus, epsilon_cross: polarization basis tensors
        ii - jj and ij + ji, respectively.

    Methods:

        getantennapatternfactors(Det) : Return F-plus and F-cross for
        Detector and GW Source objects.

    See also: Detector

    """
    def __init__(self, polarizationDeg, skyPosition=None, **kwargs):
        if (skyPosition == None):
            decDeg = kwargs.get('decDeg')
            raHr = kwargs.get('raHr')
            skyPosition = SkyPosition(decDeg,raHr)

        self.skyPosition = skyPosition
        
        self.k = skyPosition.k
        self.i = skyPosition.i
        self.j = skyPosition.j

        self.epsilon_plus = skyPosition.epsilon_plus
        self.epsilon_cross = skyPosition.epsilon_cross

        # Polarization angle in radians:
        self.psi = polarizationDeg * np.pi / 180.

    def getAntennaPatternFactors(self, detector, gstHr, *args, **kwargs):
        """Return antenna pattern factors (Fplus, Fcross).

        getantennapatternfactors(Det) returns the values for F-plus
        and F-cross for the Detector object specified. The input Det
        is a Detector object.
        
        Alternatively, F-plus and F-cross may be computed by calling
        Detector.getantennapatternfactors(Source), and specifying a
        Source object.

        Example:

        See also: Detector, Source, IFODetector, BarDetector

        """
        return detector.antennaPatterns(self, gstHr)

def LLO():
    """LIGO Livingston (Livingston, Louisiana, USA)
    
    Construct an IFODetector object for LIGO Livingston.

    """
    loc = Location(30 + (33 + 46.4196 / 60) / 60, # Latitude (degrees)
                 -(90 + (46+27.2654 / 60) / 60),  # Longitude (degrees)
                  -6.574)                         # Height (meters)
    xarm = Orientation(loc, 180+72.2835, -3.121e-4*180/np.pi)
    yarm = Orientation(loc, 180-17.7165, -6.107e-4*180/np.pi)
    return IFODetector(loc, xarm, yarm)

def LHO():
    """LIGO Hanford (Hanford, Washington, USA)

    Construct an IFODetector object for LIGO Hanford

    """
    loc  = Location( 46+(27+18.528 / 60) / 60,      # Latitude (degrees)
                   -(119+(24+27.5657 / 60) / 60),   # Longitude (degrees)
                     142.554)                   # Height (meters)
    xarm = Orientation(loc, -35.9994, -6.195e-4*180 / np.pi)
    yarm = Orientation(loc, 180+54.0006, 1.25e-5*180 / np.pi)
    return IFODetector(loc, xarm, yarm)

def VIRGO():
    """VIRGO (Cascina / Pisa, Italy)

    Construct an IFODetector object for VIRGO

    """
    loc  = Location(43 + (37 + 53.0921 / 60) / 60 ,       # Latitude (degrees)
			  10 + (30 + 16.1878 / 60) / 60 , # Longitude (degrees)
			  51.884)                     # Height (meters)
    xarm = Orientation(loc, 90-70.5674)
    yarm = Orientation(loc, 90-160.5674)
    return IFODetector(loc, xarm, yarm)

def GEO600():
    """GEO-600 (Hannover, Germany)
    
    Construct an IFODetector object for GEO-600.

    """
    loc  = Location(52 + (14 + 42.528 / 60) / 60 ,  # Latitude (degrees)
		    9 + (48 + 25.894 / 60) / 60 ,   # Longitude (degrees)
                    114.425)                    # Height (meters)
    xarm = Orientation(loc, 90-21.6117)
    yarm = Orientation(loc, 90-115.9431)
    return IFODetector(loc, xarm, yarm)

def TAMA300():
    """TAMA-300 (Tokyo, Japan)
    
    Construct an IFODetector object for TAMA-300

    """
    loc  = Location(35 + (40 + 35.6 / 60) / 60 ,  # Latitude (degrees)
		    139 + (32 + 9.8 / 60) / 60 ,  # Longitude (degrees)
	            90)                       # Height (meters)
    xarm = Orientation(loc, 90-180)
    yarm = Orientation(loc, 90-270)
    return IFODetector(loc, xarm, yarm)

def ALLEGRO(azDeg):
    """ALLEGRO (Baton Rouge, Louisiana, USA)

    Construct a BarDetector object for ALLEGRO."""
    loc = Location(30+(24+45.110 / 60) / 60,    # Latitude (degrees)
                - (91+(10+43.766 / 60) / 60) )  # Longitude (degrees)
    axis = Orientation(loc, azDeg)
    return BarDetector(loc, axis)

def AURIGA(azDeg):
    """AUnRIGA (Legnaro / Padova, Italy)

    Construct a BarDetector object for AURIGA.

    """
    loc = Location(45+(21+12 / 60) / 60, # Latitude (degrees)
                   11+(56+54 / 60) / 60) # Longitude (degrees)
    axis = Orientation(loc, azDeg)
    return BarDetector(loc, axis)

def EXPLORER(azDeg):
    """EXPLORER (Geneva, Switzerlad)

    Construct a BarDetector object for EXPLORER.

    """
    loc = Location(46+27 / 60, # Latitude (degrees)
		   6+12 / 60 ) # Longitude (degrees)
    axis = Orientation(loc, azDeg)
    return BarDetector(loc, axis)

def NAUTILUS(azDeg):
    """NAUTILUS (Frascati / Rome, Italy)
    
    Construct a BarDetector object for NAUTILUS.

    """
    loc = Location(41+(49+26 / 60) / 60,  # Latitude (degrees)
		   12+(40+21 / 60) / 60 ) # Longitude (degrees)
    axis = Orientation(loc, azDeg)
    return BarDetector(loc, axis)

def NIOBE(azDeg):
    """NIOBE (Perth, Australia)

    Construct a BarDetector object for NIOBE.

    """
    loc = Location(-(31+56 / 60), # Latitude (degrees)
		    115+49 / 60 ) # Longitude (degrees)
    axis = Orientation(loc, azDeg)
    return BarDetector(loc, axis)

site = { 'L' : LLO, 'H' : LHO, 'V' : VIRGO, 'G' : GEO600, 'T' : TAMA300, 
'A' : ALLEGRO, 'O' : AURIGA, 'E' : EXPLORER, 'N' : NAUTILUS, 'B' : NIOBE}



def overlapreductionfunction(f, Det1, Det2):
    """Return the overlap reduction function between two GW detectors.

    overlapreductionfunction calculates the overlap reduction function
    at the frequencies specified in f, a 1xn dimensional array between
    the detectors Det1 and Det2. Det1 and Det2 can be either
    IFODetector or Bardetector objects.

    The detector geometry structures can either be constructed with
    the IFODetector or BarDetector classes, or by calling a predifined
    detector in the site dictionary.

    See also: IFODetector, BarDetector, detgeom
    
    """
    LAL_C_SI = 2.99792458e+8 # m/s
    s = Det2.r - Det1.r # separation vector between sites
    distance = np.linalg.norm(s) # distance between sites in meters
    d1 = Det1.d - np.eye(3) * np.trace(Det1.d)/3 # traceless response tensor
    d2 = Det2.d - np.eye(3) * np.trace(Det1.d)/3 # traceless response tensor
    
    if (distance == 0.):
        gamma = np.ones((f.shape[0]))  * 2. * np.trace(np.dot(d1, d2))
    
    else:
        s = s/distance  # unit separation vector
        c1 = np.sum(np.sum(d1 * d2))
        c2 = np.dot(np.dot(np.dot(s, d1), d2), s)
        c3 = np.dot(np.dot (np.dot(s, d1), s), np.dot(np.dot(s, d2), s))

        alpha = f * (distance * 2 * np.pi / LAL_C_SI)
        alpha2 = alpha * alpha
     
        b0 = np.ones((alpha.shape[0]))
        b1 = b0 / 3
        b2 = b0 / 15

        i = np.nonzero(alpha)[0]

        for number in i: b0[number] = sci.sph_jn(0, alpha[number])[0]

        for number in i: b1[number] = (sci.sph_jn(1, alpha[number])[0][1] 
                                       / alpha[number])
        for number in i: b2[number] = (sci.sph_jn(2, alpha[number])[0][2] 
                                       / alpha2[number])

        rho1 = 5.0 * b0 - 10.0 * b1 +  5.0 * b2
        rho2 = -10.0 * b0 + 40.0 * b1 - 50.0 * b2
        rho3 =   2.5 * b0 - 25.0 * b1 + 87.5 * b2

        gamma = c1 * rho1 + c2 * rho2 + c3 * rho3

        return gamma

def orfintegrand(f, Det1, Det2, n):
    LAL_C_SI = 2.99792458e+8

    d1 = Det1.d - np.eye(3) * np.trace(Det1.d)/3 # traceless response tensor
    d2 = Det2.d - np.eye(3) * np.trace(Det1.d)/3 # traceless response tensor

    profreq = np.ones((np.shape(f)))
    sizedir = np.shape(n)
    prodir = np.ones((1, sizedir[0]))

    d1n = np.dot(d1, n)
    d2n = np.dot(d2, n)

    c1 = np.sum(np.sum(d1 * d2))
    c2 = np.sum(d1n * d2n)
    c3 = np.sum(n * d1n) * np.sum(n * d2n)

    magnitude = np.dot(profreq * (5 / 4 * np.pi) , (c1 - 2 * c2 + c3 / 2))
    phase = (2 * np.pi / LAL_C_SI) * f * np.dot((Det2.r - Det1.r), n)

    gamma = magnitude * np.exp(1j * phase)

    return gamma

def getcartesiandirectionfromsource(decDeg, minusHAHr):
    """ Return Cartesian unit vector for wave propagation from declination and 
    minus hour angle.

    """
    cosdelta = np.cos(decDeg * np.pi / 180);
  
    n = np.array([-cosdelta * np.cos(minusHAHr * np.pi / 12), 
                  -cosdelta * np.sin(minusHAHr * np.pi / 12), 
                  -np.sin(decDeg * np.pi / 180)])
    return n

def getGmstHrFromGps(gps):
    """ Return Greenwich Mean Sidereal Time in hours, given GPS time
    in seconds.  Note that for converting between corotating to
    inertial coordinates, you should use getGstHrFromGps() instead

    This implements the formulas at

    http://www.usno.navy.mil/USNO/astronomical-applications/astronomical-information-center/approx-sider-time
    """

    # GPS time of reference epoch, obtained by
    # lalapps_tconvert 'Sat Jan 01 12:00:00 GMT 2000'
    gps0 = 630763213.
    # Days since epoch
    D = (gps-gps0)/86400.
    # Julian Day of epoch
    JDEpoch = 2451545.0

    # Correct for leap seconds to get more accurate accounting of time
    # elapsed.  Behavior may be unexpected exactly at the leap second
    LSDArray = leapSecs - JDEpoch
    leapSecOffset = 0

    if D > 0:
        while (( (D - leapSecOffset/86400. > LSDArray)
                 & (LSDArray > 0) ).sum()
               > leapSecOffset) :
            leapSecOffset = ( (D - leapSecOffset/86400. > LSDArray)
                              & (LSDArray > 0) ).sum()
    elif D < 0:
        while (( (D - leapSecOffset/86400. < LSDArray)
                 & (LSDArray < 0) ).sum()
               > -leapSecOffset) :
            leapSecOffset = -( (D - leapSecOffset/86400. < LSDArray)
                               & (LSDArray < 0) ).sum()

    D -= leapSecOffset/86400.

    # Centuries since epoch
    T = D/36525.
    # The equation on the USNO website is
    #
    # GMST = 6.697374558 + 0.06570982441908 D0 + 1.00273790935 H + 0.000026 T**2
    #
    # But since D = D0 + H/24 and 24.06570982441908/24 = 1.00273790935
    # this is the same as
    #
    # GMST = 6.697374558 + 24.06570982441908 D - 24 D0 + 0.000026 T**2
    #
    # and since D0 is guaranteed to be a half-integer, this is the
    # same thing, modulo 24, as
    GMST = 18.697374558 + 24.06570982441908 * D + 0.000026 * T**2

    return GMST % 24

def getGstHrFromGps(gps):
    """ Return Greenwich Sidereal Time in hours, given GPS time
    in seconds.

    This implements the formulas at

    http://www.usno.navy.mil/USNO/astronomical-applications/astronomical-information-center/approx-sider-time
    """

    # GPS time of reference epoch, obtained by
    # lalapps_tconvert 'Sat Jan 01 12:00:00 GMT 2000'
    gps0 = 630763213.
    # Days since epoch
    D = (gps-gps0)/86400.

    # Greenwich mean sidereal time
    GMST = getGmstHrFromGps(gps)

    # Radians per degree
    radPerDeg = np.pi/180.

    # Longitude of the ascending node of the Moon
    OmegaDeg = 125.04 - 0.052954 * D

    # Mean Longitude of the Sun
    LDeg = 280.47 + 0.98565 * D

    # obliquity
    epsDeg = 23.4393 - 0.0000004 * D

    # nutation in longitude
    delPsiHr = - 0.000319 * np.sin(OmegaDeg * radPerDeg) \
        - 0.000024 * np.sin(2. * LDeg * radPerDeg)

    # equation of the equinoxes
    eqeq = delPsiHr * np.cos(epsDeg * radPerDeg)

    GAST = GMST + eqeq

    return GAST % 24

def getDmsFromDecimal(decimal):
    d = int(decimal)
    minutes = 60*(decimal-d)
    m = int(minutes)
    seconds = 60*(minutes-m)
    return (d,m,seconds)

def getDecimalFromDms(d,m,seconds):
    m = np.sign(d) * np.abs(m)
    seconds = np.sign(d) * np.abs(seconds)
    return d + m/60. + seconds/3600.
